"""This subpackage contains `PipelineItem`s used to interact with a
[TFaaS inference server](https://github.com/vkuznet/TFaaS/).
"""

# from CHAP.inference.reader import
from CHAP.inference.processor import TFaaSImageProcessor
# from CHAP.inference.writer import
