"""This subpackage contains a collection of
[pydantic](https://docs.pydantic.dev/latest/) models that are used by
`Processors` in [`CHAP.common`](CHAP.common.md) and elsewhere to
validate inputs.
"""
